/************************************************************************
 *                                                                      *
 *  FreeSynd - a remake of the classic Bullfrog game "Syndicate".       *
 *                                                                      *
 *   Copyright (C) 2005  Stuart Binge  <skbinge@gmail.com>              *
 *   Copyright (C) 2005  Joost Peters  <joostp@users.sourceforge.net>   *
 *   Copyright (C) 2006  Trent Waddington <qg@biodome.org>              *
 *                                                                      *
 *    This program is free software;  you can redistribute it and / or  *
 *  modify it  under the  terms of the  GNU General  Public License as  *
 *  published by the Free Software Foundation; either version 2 of the  *
 *  License, or (at your option) any later version.                     *
 *                                                                      *
 *    This program is  distributed in the hope that it will be useful,  *
 *  but WITHOUT  ANY WARRANTY;  without even  the implied  warranty of  *
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 *  General Public License for more details.                            *
 *                                                                      *
 *    You can view the GNU  General Public License, online, at the GNU  *
 *  project's  web  site;  see <http://www.gnu.org/licenses/gpl.html>.  *
 *  The full text of the license is also included in the file COPYING.  *
 *                                                                      *
 ************************************************************************/

#include <cstdio>

#include "common.h"
#include "fliplayer.h"
#include "screen.h"
#include "app.h"

FliPlayer::~FliPlayer() {
    if (offscreen_)
        delete[] offscreen_;
}

void FliPlayer::loadFliData(uint8 *data, bool skipable) {
    fli_data_ = data;
    skipable_ = skipable;

    fli_info_.size           = READ_LE_UINT32(fli_data_); fli_data_ += 4;
    fli_info_.type           = READ_LE_UINT16(fli_data_); fli_data_ += 2;
    fli_info_.numFrames      = READ_LE_UINT16(fli_data_); fli_data_ += 2;
    fli_info_.width          = READ_LE_UINT16(fli_data_); fli_data_ += 2;
    fli_info_.height         = READ_LE_UINT16(fli_data_); fli_data_ += 2;

    if (fli_info_.type != 0xAF12) {  //simple check to verify it is indeed a (Bullfrog) FLI
        printf("FliPlayer::loadFliData(): attempted to load non-FLI data (type = 0x%04X)\n", fli_info_.type);
        fli_info_.width = fli_info_.height = 100;
        fli_info_.numFrames = 0;
        return;
    }

    assert(fli_info_.width == 320 && fli_info_.height == 200);
    if (offscreen_)
        delete[] offscreen_;
    offscreen_ = new uint8[fli_info_.width * fli_info_.height];

    memset(palette_, 0, sizeof(palette_));
}

bool FliPlayer::isValidChunk(uint16 type) {
    //Even though it may be a valid chunk type, only return true if we know how to deal with it
    switch (type) {
        case         4: //COLOR_256
        case         7: //DELTA_FLC (FLI_SS2)
        case        15: //BYTE_RUN
        case    0xF1FA: //FRAME_TYPE
                return true;

        default:
            printf("ERROR: isValidChunk(0x%04X) is NOT a valid chunk\n", type);
            return false;
    }
}

ChunkHeader FliPlayer::readChunkHeader(uint8 *mem) {
    ChunkHeader head;
    head.size = READ_LE_UINT32(mem + 0);
    head.type = READ_LE_UINT16(mem + 4);
    return head;
}

FrameTypeChunkHeader FliPlayer::readFrameTypeChunkHeader(ChunkHeader chunkHead, uint8 *&mem) {
    FrameTypeChunkHeader head;

    head.header = chunkHead; mem += 6;
    head.numChunks = READ_LE_UINT16(mem); mem += 2;
    head.delay = READ_LE_UINT16(mem); mem += 2;
    head.reserved = READ_LE_UINT16(mem); mem += 2;
    head.widthOverride = READ_LE_UINT16(mem); mem += 2;
    head.heightOverride = READ_LE_UINT16(mem); mem += 2;

    return head;
}

void FliPlayer::decodeByteRun(uint8 *data) {
    uint8 *ptr = (uint8 *)offscreen_;
    while((ptr - offscreen_) < (fli_info_.width * fli_info_.height)) {
        uint8 chunks = *data++;
        while (chunks--) {
            int8 count = *data++;
            if (count > 0) {
                while (count--) {
                    *ptr++ = *data;
                }
                data++;
            } else {
                uint8 copyBytes = -count;
                memcpy(ptr, data, copyBytes);
                ptr += copyBytes;
                data += copyBytes;
            }
        }
    }
}

#define OP_PACKETCOUNT      0
#define OP_UNDEFINED        1
#define OP_LASTPIXEL        2
#define OP_LINESKIPCOUNT    3

void FliPlayer::decodeDeltaFLC(uint8 *data) {
    uint16 linesInChunk = READ_LE_UINT16(data); data += 2;
    uint16 currentLine = 0;
    uint16 packetCount = 0;

    while (linesInChunk--) {
        uint16 opcode;

        // First process all the opcodes.
        do {
            opcode = READ_LE_UINT16(data); data += 2;

            switch ((opcode >> 14) & 3) {
                case    OP_PACKETCOUNT:
                        packetCount = opcode;
                        break;
                case    OP_UNDEFINED:
                        break;
                case    OP_LASTPIXEL:
                        *(uint8 *)(offscreen_ + (currentLine * fli_info_.width) + (fli_info_.height - 1)) = (opcode & 0xFF);
                        break;
                case    OP_LINESKIPCOUNT:
                        currentLine += -(int16)opcode;
                        break;
            }
        } while (((opcode >> 14) & 3) != OP_PACKETCOUNT);

        uint16 column = 0;

        //Now interpret the RLE data
        while (packetCount--) {
            column += *data++;
            int8 rleCount = (int8)*data++;

            if (rleCount > 0) {
                memcpy((void *)(offscreen_ + (currentLine * fli_info_.width) + column), data, rleCount * 2);
                data += rleCount * 2;
                column += rleCount * 2;
            } else if (rleCount < 0) {
                uint16 dataWord = *(uint16 *)data; data += 2;
                for (int i = 0; i < -(int16)rleCount; ++i) {
                    *(uint16 *)(offscreen_ + (currentLine * fli_info_.width) + column + (i * 2)) = dataWord;
                }

                column += (-(int16)rleCount) * 2;
            } else {    // End of cutscene ?
                return;
            }
        }

        currentLine++;
    }
}


#define FRAME_TYPE  0xF1FA

bool FliPlayer::decodeFrame() {
    FrameTypeChunkHeader frameHeader;
    ChunkHeader cHeader = readChunkHeader(fli_data_);
    do {
        switch(cHeader.type) {
            case    4:
                    setPalette(fli_data_ + 6);
                    g_System.setPalette8b3(palette_);
                    break;
            case    7:
                    decodeDeltaFLC(fli_data_ + 6);
                    break;
            case    15:
                    decodeByteRun(fli_data_ + 6);
                    break;
            case    FRAME_TYPE:
                    frameHeader = readFrameTypeChunkHeader(cHeader, fli_data_);
                    fli_info_.numFrames--;
                    //printf("Frames Remaining: %d\n", fli_info_.numFrames);
                    break;
            default:
                    break;
         }

        if (cHeader.type != FRAME_TYPE)
            fli_data_ += cHeader.size;

        cHeader = readChunkHeader(fli_data_);

    } while(isValidChunk(cHeader.type) && cHeader.type != FRAME_TYPE);

    return isValidChunk(cHeader.type);

}

void FliPlayer::setPalette(uint8 *mem) {
    uint16 numPackets = READ_LE_UINT16(mem); mem += 2;

    if (0 == READ_LE_UINT16(mem)) { //special case
        mem += 2;
        for (int i = 0; i < 256; ++i)
            for (int j = 0; j < 3; ++j)
                palette_[i * 3 + j] = (mem[i * 3 + j] << 2) | (mem[i * 3 + j] & 3);
    } else {
        uint8 palPos = 0;

        while (numPackets--) {
            palPos += *mem++;
            uint8 change = *mem++;

            for (int i = 0; i < change; ++i)
                for (int j = 0; j < 3; ++j)
                    palette_[(palPos + i) * 3 + j] = (mem[i * 3 + j] << 2) | (mem[i * 3 + j] & 3);

            palPos += change;
            mem += (change * 3);
        }
    }
}

void FliPlayer::copyCurrentFrameToScreen()
{
    g_Screen.scale2x(0, 0, fli_info_.width, fli_info_.height, offscreen(), 0, false);
}

bool FliPlayer::play()
{
    if (!fli_data_)
        return false;

    g_Screen.clear(0);
    g_App.startedPlayingFli();
    while (hasFrames() && g_App.isRunning() && (!g_App.skipFli() || !skipable_)) {
        g_System.handleEvents();

        if (!decodeFrame()) break;
        copyCurrentFrameToScreen();

        g_System.updateScreen();
        g_System.delay(1000 / 15);  //15 fps
    }

    g_App.finishedPlayingFli();

    //clear the backscreen
    //bzero(Screen::pixels(), GAME_SCREEN_WIDTH * GAME_SCREEN_HEIGHT);

    return true;
}

