/************************************************************************
 *                                                                      *
 *  FreeSynd - a remake of the classic Bullfrog game "Syndicate".       *
 *                                                                      *
 *   Copyright (C) 2005  Stuart Binge  <skbinge@gmail.com>              *
 *   Copyright (C) 2005  Joost Peters  <joostp@users.sourceforge.net>   *
 *   Copyright (C) 2006  Trent Waddington <qg@biodome.org>              *
 *   Copyright (C) 2006  Tarjei Knapstad <tarjei.knapstad@gmail.com>    *
 *                                                                      *
 *    This program is free software;  you can redistribute it and / or  *
 *  modify it  under the  terms of the  GNU General  Public License as  *
 *  published by the Free Software Foundation; either version 2 of the  *
 *  License, or (at your option) any later version.                     *
 *                                                                      *
 *    This program is  distributed in the hope that it will be useful,  *
 *  but WITHOUT  ANY WARRANTY;  without even  the implied  warranty of  *
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 *  General Public License for more details.                            *
 *                                                                      *
 *    You can view the GNU  General Public License, online, at the GNU  *
 *  project's  web  site;  see <http://www.gnu.org/licenses/gpl.html>.  *
 *  The full text of the license is also included in the file COPYING.  *
 *                                                                      *
 ************************************************************************/

#include <assert.h>
#include "screen.h"
#include "sprite.h"

void unpackBlocks1(const uint8 * data, uint8 * pixels)
{
    for (int i = 0; i < 8; ++i) {
        if (bitSet(data[0], 7 - i)) {
            pixels[i] = 255;    // transparent
        } else {
            pixels[i] =
                  static_cast<uint8>((bitValue(data[1], 7 - i) << 0) & 0xff)
                | static_cast<uint8>((bitValue(data[2], 7 - i) << 1) & 0xff)
                | static_cast<uint8>((bitValue(data[3], 7 - i) << 2) & 0xff)
                | static_cast<uint8>((bitValue(data[4], 7 - i) << 3) & 0xff);
        }
    }
}

Sprite::Sprite()
  : width_(0)
  , height_(0)
  , stride_(0)
  , sprite_data_(0)
{
}

Sprite::~Sprite()
{
    if (sprite_data_)
        delete [] sprite_data_;

    width_ = height_ = stride_ = 0;
    sprite_data_ = 0;
}

bool Sprite::loadSprite(uint8 * tabData, uint8 * spriteData, uint32 offset, bool rle)
{
    assert(tabData);
    assert(spriteData);

    uint8 * tabEntry = tabData + offset * TABENTRY_SIZE;

    uint32 spriteOffset = READ_LE_UINT32(tabEntry);
    tabEntry += 4;
    width_ = *tabEntry;
    tabEntry += 1;
    height_ = *tabEntry;

    if (width_ == 0 || height_ == 0)
        return true;

    stride_ = ceil8(width_);
    uint8 * spriteBlocks = spriteData + spriteOffset;

    sprite_data_ = new uint8 [stride_ * height_];
    memset(sprite_data_, 255, stride_ * height_);

    uint8 * currentPixel;

    if (rle) {
        for (int i = 0; i < height_; ++i) {
            int spriteWidth = width_;
            currentPixel = sprite_data_ + i * stride_;

            int runLength = static_cast<char>(*spriteBlocks++);
            while (runLength != 0) {
                spriteWidth -= runLength;

                if (runLength > 0) {
                // pixel run
                    for (int j = 0; j < runLength; ++j)
                        *currentPixel++ = *spriteBlocks++;

                } else if (runLength < 0) {
                // transparent run
                    runLength *= -1;
                    memset(currentPixel, 255, runLength);
                    currentPixel += runLength;

                } else if (runLength == 0) {
                // end of the row
                    spriteWidth = 0;
                }

                runLength = static_cast<char>(*spriteBlocks++);
            }
        }
    } else {
        for (int j = 0; j < height_; ++j) {
            currentPixel = sprite_data_ + j * stride_;

            for (int i = 0; i < width_; i += PIXELS_PER_BLOCK) {
                unpackBlocks1(spriteBlocks, currentPixel);

                spriteBlocks += BLOCK_LENGTH;
                currentPixel += PIXELS_PER_BLOCK;
            }
        }
    }

    return true;
}

void Sprite::draw(int x, int y, int z, bool flipped, bool x2)
{
    if (x2)
	g_Screen.scale2x(x, y, width_, height_, sprite_data_, stride_);
    else
	g_Screen.blit(x, y, width_, height_, sprite_data_, flipped, stride_);
}

void Sprite::data(uint8 *spr_data) const
{
    for (int j = 0; j < height_; j++) {
	memcpy(spr_data + j * width_, sprite_data_ + j * stride_, width_);
    }
}
