/************************************************************************
 *                                                                      *
 *  FreeSynd - a remake of the classic Bullfrog game "Syndicate".       *
 *                                                                      *
 *   Copyright (C) 2005  Stuart Binge  <skbinge@gmail.com>              *
 *   Copyright (C) 2005  Joost Peters  <joostp@users.sourceforge.net>   *
 *   Copyright (C) 2006  Trent Waddington <qg@biodome.org>              *
 *                                                                      *
 *    This program is free software;  you can redistribute it and / or  *
 *  modify it  under the  terms of the  GNU General  Public License as  *
 *  published by the Free Software Foundation; either version 2 of the  *
 *  License, or (at your option) any later version.                     *
 *                                                                      *
 *    This program is  distributed in the hope that it will be useful,  *
 *  but WITHOUT  ANY WARRANTY;  without even  the implied  warranty of  *
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 *  General Public License for more details.                            *
 *                                                                      *
 *    You can view the GNU  General Public License, online, at the GNU  *
 *  project's  web  site;  see <http://www.gnu.org/licenses/gpl.html>.  *
 *  The full text of the license is also included in the file COPYING.  *
 *                                                                      *
 ************************************************************************/

#include <stdio.h>
#include <assert.h>
#include "app.h"
#include "file.h"

SpriteManager::SpriteManager() : sprites_(0), sprite_count_(0)
{
}

SpriteManager::~SpriteManager()
{
    clear();
}

void SpriteManager::clear()
{
    if (sprites_)
        delete [] sprites_;

    sprites_ = 0;
    sprite_count_ = 0;
}

bool SpriteManager::loadSprites(uint8 * tabData, int tabSize, uint8 * spriteData, bool rle)
{
    assert(tabData);
    assert(spriteData);

    sprite_count_ = tabSize / TABENTRY_SIZE;
    sprites_ = new Sprite [sprite_count_];
    assert(sprites_);

    for (int i = 0; i < sprite_count_; ++i) {
        if (!sprites_[i].loadSprite(tabData, spriteData, i, rle)) {
            printf("Failed to load sprite: %d\n", i);
        }
    }

    return true;
}

Sprite *SpriteManager::sprite(int spriteNum)
{
    if (spriteNum >= sprite_count_) {
        printf("SpriteManager::drawSpriteXYZ: sprite %d not loaded!\n", spriteNum);
        return NULL;
    }
    return &sprites_[spriteNum];
}


bool SpriteManager::drawSpriteXYZ(int spriteNum, int x, int y, int z, bool flipped, bool x2)
{
    if (spriteNum >= sprite_count_) {
        printf("SpriteManager::drawSpriteXYZ: sprite %d not loaded!\n", spriteNum);
        return false;
    }

    sprites_[spriteNum].draw(x, y, z, flipped, x2);

    return true;
}


GameSpriteManager::GameSpriteManager()
{
}

GameSpriteManager::~GameSpriteManager()
{
}

void GameSpriteManager::load()
{
    int tabSize, size;
    uint8 *tabData, *data;
#if 1
    // using http://rtfm.insomnia.org/~qg/syndicate.zip
    tabData = File::loadFile("hspr-0.tab", tabSize);
    data = File::loadFile("hspr-0.dat", size);
    printf("Loaded %d sprites from hspr-0.dat\n", tabSize / 6);
    g_App.gameSprites().loadSprites(tabData, tabSize, data);
    delete [] tabData;
    delete [] data;
#else
    tabData = File::loadFile("hspr-0-d.tab", tabSize);
    data = File::loadFile("hspr-0-d.dat", size);
    printf("Loading %d sprites from hspr-0-d.dat\n", tabSize / 6);
    g_App.gameSprites().loadSprites(tabData, tabSize, data);
    delete [] tabData;
    delete [] data;
#endif

    data = File::loadFile("HELE-0.ANI", size);

    assert(size % 10 == 0);
    for (int i = 0; i < size / 10; i++) {
        GameSpriteFrameElement e;
        e.sprite_ = data[i * 10] | (data[i * 10 + 1] << 8);
        assert(e.sprite_ % 6 == 0);
        e.sprite_ /= 6;
        e.off_x_ = data[i * 10 + 2] | (data[i * 10 + 3] << 8);
        e.off_y_ = data[i * 10 + 4] | (data[i * 10 + 5] << 8);
        e.flipped_ = (data[i * 10 + 6] | (data[i * 10 + 7] << 8)) != 0 ? true : false;
        e.next_element_ = data[i * 10 + 8] | (data[i * 10 + 9] << 8);
        if (e.off_x_ & (1 << 15))
            e.off_x_ = -(65536 - e.off_x_);
        if (e.off_y_ & (1 << 15))
            e.off_y_ = -(65536 - e.off_y_);
        assert(e.next_element_ < size / 10);
        elements_.push_back(e);
    }
    delete[] data;

    printf("loaded %i frame elements\n", size / 10);

    data = File::loadFile("HFRA-0.ANI", size);

    assert(size % 8 == 0);
    for (int i = 0; i < size / 8; i++) {
        GameSpriteFrame f;
        f.first_element_ = data[i * 8] | (data[i * 8 + 1] << 8);
        assert(f.first_element_ < (int)elements_.size());
        f.width_ = data[i * 8 + 2];
        f.height_ = data[i * 8 + 3];
        f.flags_ = data[i * 8 + 4] | (data[i * 8 + 5] << 8);
        f.next_frame_ = data[i * 8 + 6] | (data[i * 8 + 7] << 8);
        assert(f.next_frame_ < size / 8);
        frames_.push_back(f);
    }
    delete[] data;

    printf("loaded %i frames\n", size / 8);

    data = File::loadFile("HSTA-0.ANI", size);

    assert(size % 2 == 0);
    for (int i = 0; i < size / 2; i++) {
        index_.push_back(data[i * 2] | (data[i * 2 + 1] << 8));
        assert(index_[i] < (int)frames_.size());
    }
    delete[] data;

    printf("index contains %i animations\n", size/2);
}

bool GameSpriteManager::drawFrame(int animNum, int frameNum, int x, int y)
{
    assert(animNum < (int)index_.size());

    GameSpriteFrame *f = &frames_[index_[animNum]];
    while (frameNum) {
        f = &frames_[f->next_frame_];
        frameNum--;
    }

    GameSpriteFrameElement *e = &elements_[f->first_element_];
    while(1) {
        sprites_[e->sprite_].draw(x + e->off_x_, y + e->off_y_, 0, e->flipped_);
        if (e->next_element_ == 0)
            break;
        e = &elements_[e->next_element_];
    } 

    return f->next_frame_ == index_[animNum];
}

bool GameSpriteManager::lastFrame(int animNum, int frameNum)
{
    assert(animNum < (int)index_.size());

    GameSpriteFrame *f = &frames_[index_[animNum]];
    while (frameNum) {
        f = &frames_[f->next_frame_];
        frameNum--;
    }

    return f->next_frame_ == index_[animNum];
}

int GameSpriteManager::lastFrame(int animNum)
{
    int frameNum = 0;
    assert(animNum < (int)index_.size());

    GameSpriteFrame *f = &frames_[index_[animNum]];
    while (1) {
        f = &frames_[f->next_frame_];
	if (f->next_frame_ == index_[animNum])
	    return frameNum;
        frameNum++;
    }
}
