/************************************************************************
 *                                                                      *
 *  FreeSynd - a remake of the classic Bullfrog game "Syndicate".       *
 *                                                                      *
 *   Copyright (C) 2005  Stuart Binge  <skbinge@gmail.com>              *
 *   Copyright (C) 2005  Joost Peters  <joostp@users.sourceforge.net>   *
 *   Copyright (C) 2006  Trent Waddington <qg@biodome.org>              *
 *   Copyright (C) 2006  Tarjei Knapstad <tarjei.knapstad@gmail.com>    *
 *                                                                      *
 *    This program is free software;  you can redistribute it and / or  *
 *  modify it  under the  terms of the  GNU General  Public License as  *
 *  published by the Free Software Foundation; either version 2 of the  *
 *  License, or (at your option) any later version.                     *
 *                                                                      *
 *    This program is  distributed in the hope that it will be useful,  *
 *  but WITHOUT  ANY WARRANTY;  without even  the implied  warranty of  *
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 *  General Public License for more details.                            *
 *                                                                      *
 *    You can view the GNU  General Public License, online, at the GNU  *
 *  project's  web  site;  see <http://www.gnu.org/licenses/gpl.html>.  *
 *  The full text of the license is also included in the file COPYING.  *
 *                                                                      *
 ************************************************************************/

#include "xmidi.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>

#define XMIDI_CONTROLLER_NEXT_BREAK     117

std::vector<XMidi::Midi> XMidi::convertXMidi(uint8* buf, int size)
{
    std::vector<Midi> midi;
    XMidiFile* xmi = loadXMidi(buf, size);
    if (!xmi)
    {
	return midi;
    }
    midi.resize(xmi->tracks);
    for(unsigned int i = 0; i < xmi->tracks; ++i)
    {
	midi[i].size_ = retrieveMidi(xmi, i, &(midi[i].data_));
    }
    return midi;
}

XMidi::XMidiFile* XMidi::loadXMidi(uint8* buf, int size)
{
    XMidiFile	*xmidi = NULL;
    xmidi = (XMidiFile*)malloc (sizeof(XMidiFile));
    xmidi->events = NULL;
    xmidi->timing = NULL;

    if (!readFile (xmidi, buf))
    {
	    freeXMidi(xmidi);
	    xmidi = NULL;
    }
    return xmidi;
}

void XMidi::freeXMidi(XMidiFile *xmidi)
{
    int i;
    if (xmidi->events)
    {
	for (i=0; i < xmidi->tracks; i++)
	    deleteEventList (xmidi->events[i]);
	free(xmidi->events);
    }
    if (xmidi->timing) free(xmidi->timing);
}

int XMidi::retrieveMidi(XMidiFile *xmidi, uint32 track, uint8 **buffer)
{
    int len = 0;
    uint8 *buf;
    
    if (!xmidi->events)
    {
	fprintf (stderr, "No midi data in loaded\n");
	return 0;
    }
    
    if (track >= xmidi->tracks)
    {
	fprintf (stderr, "Can't retrieve MIDI data, track out of range\n");
	return 0;
    }
    
    len = 14 + convertListToMTrk (NULL, xmidi->events[track]);
    
    if (len == 14)
    {
	len = 0;
	return 0;
    }
    
    buf = *buffer = (unsigned char*)malloc(len);
    
    if (!buf)
    {
	fprintf (stderr, "Unable to allocate output buffer\n");
	return 0;
    }
    
    buf[0] = 'M';
    buf[1] = 'T';
    buf[2] = 'h';
    buf[3] = 'd';
    
    buf[4] = 0;
    buf[5] = 0;
    buf[6] = 0;
    buf[7] = 6;
    
    buf[8] = 0;
    buf[9] = 0;
    
    buf[10] = 0;
    buf[11] = 1;
    
    buf[12] = (unsigned char) (xmidi->timing[track] >> 8);
    buf[13] = (unsigned char) (xmidi->timing[track] & 0xFF);
    
    convertListToMTrk (buf+14, xmidi->events[track]);
    
    return len;
}



XMidi::midi_event* XMidi::allocateEvent(int time)
{
  midi_event *event = (midi_event*)malloc (sizeof (midi_event));
	event->status = 0;
	event->next = NULL;
	event->time = time;
	event->stream = NULL;
	event->len = 0;

	return event;
}

/* Sets current to the new event and updates list */
void XMidi::createNewEvent (XMidiFile *xmidi, int time)
{
	if (!xmidi->list)
	{
		xmidi->list = xmidi->current = allocateEvent(time);
		return;
	}

	if (xmidi->current->time > time)
		xmidi->current = xmidi->list;

	while (xmidi->current->next)
	{
		if (xmidi->current->next->time > time)
		{
			midi_event *event = allocateEvent(time);
			
			event->next = xmidi->current->next;
			xmidi->current->next = event;
			xmidi->current = event;
			return;
		}
		
		xmidi->current = xmidi->current->next;
	}

	xmidi->current->next = allocateEvent(time);
	xmidi->current = xmidi->current->next;
}


int XMidi::getVLQ (const unsigned char *stream, uint32 *quant)
{
	int i;
	*quant = 0;
	
	for (i = 0; i < 4; i++)
	{
		*quant <<= 7;
		*quant |= stream[i] & 0x7F;
		
		if (!(stream[i] & 0x80))
		{
			i++;
			break;
		}
	}

	return i;
}

int XMidi::getVLQ2 (const unsigned char *stream, uint32 *quant)
{
	int i;
	*quant = 0;
	
	for (i = 0; i < 4 && !(stream[i] & 0x80); i++)
		*quant += stream[i];
	
	return i;
}

int XMidi::putVLQ(unsigned char *stream, uint32 value)
{
	int buffer;
	int i = 1;
	buffer = value & 0x7F;
	while (value >>= 7)
	{
		buffer <<= 8;
		buffer |= ((value & 0x7F) | 0x80);
		i++;
	}
	if (!stream) return i;
	while (1)
	{
		*stream++ = (buffer & 0xFF);
		if (buffer & 0x80)
			buffer >>= 8;
		else
			break;
	}
	
	return i;
}

/*
// Converts Events
//
// Pointer to stream is at the status byte
// size 1 is single data byte
// size 2 is dual data byte
// size 3 is XMI Note on
// Returns bytes converted
*/

int XMidi::convertEvent (XMidiFile *xmidi, const int time, const unsigned char status, const unsigned char *stream, const int size)
{
	int i;
	uint32 delta = 0;
	midi_event *prev;
	
	createNewEvent (xmidi, time);
	xmidi->current->status = status;
	xmidi->current->data[0] = stream[0];

	if (size == 1) return 1;

	xmidi->current->data[1] = stream[1];

	if (size == 2) return 2;

	/* Save old */
	prev = xmidi->current;

	i = getVLQ (stream+2, &delta);

	createNewEvent (xmidi, time+delta*3);

	xmidi->current->status = status;
	xmidi->current->data[0] = stream[0];
	xmidi->current->data[1] = 0;
	
	/* Change the xmidi->current to the prev */
	xmidi->current = prev;

	return i + 2;
}

/* Simple routine to convert system messages */
int XMidi::convertSystemMessage (XMidiFile *xmidi, const int time, const unsigned char *stream)
{
	int i=1;
	
	createNewEvent (xmidi, time);
	xmidi->current->status = stream[0];
	
	/* Handling of Meta events */
	if (stream[0] == 0xFF)
	{
		xmidi->current->data[0] = stream[1];
		i++;	
	}

	i += getVLQ (stream+i, &xmidi->current->len);

	if (!xmidi->current->len) return i;

	xmidi->current->stream = (unsigned char*)malloc(xmidi->current->len);

	memcpy (xmidi->current->stream, stream+i, xmidi->current->len);

	return i+xmidi->current->len;
}

/*
// XMidiFile to List
// Returns PPQN 
*/
int XMidi::readEventList (XMidiFile *xmidi, const unsigned char *stream)
{
	int	time = 0;
	uint32 delta;
	int	end = 0;
	int	tempo = 500000;
	int	tempo_set = 0;
	
	int i = 0;
	
	for (i = 0; !end; )
	{
		switch (stream[i] >> 4)
		{
			/* Note Off */
			case 0x8:
			fprintf (stderr, "ERROR: Note off not valid in XMidiFile\n");
/*			deleteEventList (xmidi->list); */
/*			xmidi->list = NULL; */
			return 0;
			
			
			/* Note On */
			case 0x9:
			i+= 1+convertEvent (xmidi, time, stream[i], stream+i+1, 3);
			break;

			/* 2 byte data */
			/* Aftertouch, Controller and Pitch Wheel */
			case 0xA: case 0xB: case 0xE:
			i+= 1+convertEvent (xmidi, time, stream[i], stream+i+1, 2);
			break;
			

			/* 1 byte data */
			/* Program Change and Channel Pressure */
			case 0xC: case 0xD:
			i+= 1+convertEvent (xmidi, time, stream[i], stream+i+1, 1);
			break;
			

			/* SysEx */
			case 0xF:
			if (stream[i] == 0xFF)
			{
				if (stream[i+1] == 0x2F) /* End */
					end = 1;
				else if (stream[i+1] == 0x51 && !tempo_set) /* Tempo. Need it for PPQN */
				{
					tempo = stream[i+5];
					tempo |= stream[i+4] << 8;
					tempo |= stream[i+3] << 16;
					tempo_set = 1;
				}
			}
			i+= convertSystemMessage (xmidi, time, stream+i);
			break;


			/* Delta T */
			default:
			i+= getVLQ2 (stream+i, &delta);
			delta *= 3;
			time += delta;
			break;
		}

	}
	
	return (tempo*9)/25000;
}

int XMidi::extractTracks (XMidiFile *xmidi, const unsigned char *stream, const uint32 size)
{
	uint32		i = 0;
	int		num = 0;
	signed short	ppqn;
	
	uint32	len = 0;

	while (i < size && num != xmidi->tracks)
	{
		/* Allign */
		i+= (len+1)&~1;

		/* Skip the FORM entries */
		if (!memcmp(stream+i,"FORM",4))
			i+= 12;

		len = stream[i+7];
		len |= stream[i+6] << 8;
		len |= stream[i+5] << 16;
		len |= stream[i+4] << 24;

		i+= 8;

		if (memcmp(stream+i-8,"EVNT",4))
			continue;

		xmidi->list = NULL;

		/* Convert it */
		if (!(ppqn = readEventList (xmidi, stream+i)))
		{
			fprintf (stderr, "Unable to convert data\n");
			break;
		}

		xmidi->timing[num] = ppqn;
		xmidi->events[num] = xmidi->list;
	
		/* Increment Counter */
		num++;
	}


	/* Return how many were converted */
	return num;
}

int XMidi::readFile (XMidiFile *xmidi, const unsigned char *stream)
{
	uint32	i = 0;
	int	j = 0;
	uint32	len;
	uint32	chunk_len;
	int 	count;
	
	/* Could be XMidiFile */
	if (!memcmp (stream, "FORM", 4))
	{
		/* XDIRless XMidiFile */
		if (!memcmp (stream+8, "XMID", 4))
		{	
			fprintf (stderr, "Warning: Xmidi might not be valid\n");
			xmidi->tracks = 1;
			i+=4;
			
		} /* Not an XMidiFile */
		else if (memcmp (stream+8, "XDIR", 4))
		{	
			fprintf (stderr, "Not a XMID\n");
			return 0;
			
		} /* Seems Valid */
		else 
		{
			/* Is an XMidiFile */
			/* Assume correct length?? */
			/* No way! */
			
			len = stream[7];
			len |= stream[6] << 8;
			len |= stream[5] << 16;
			len |= stream[4] << 24;
			
			xmidi->tracks = 0;
		
			for (i = 12; i < len; i++)
			{
				chunk_len = stream[i+3];
				chunk_len |= stream[i+2] << 8;
				chunk_len |= stream[i+1] << 16;
				chunk_len |= stream[i] << 24;
			
				i+=8;
				
				if (memcmp (stream+i-8, "INFO", 4))
				{	
					/* Must allign */
					i+= (chunk_len+1)&~1;
					continue;
				}

				/* Must be at least 2 bytes long */
				if (chunk_len < 2)
					break;
				
				xmidi->tracks = stream[i];
				xmidi->tracks |= stream[i+1] << 8;
			
				break;
			}
		
			/* Didn't get to fill the header */
			if (xmidi->tracks == 0)
			{
				fprintf (stderr, "Not a valid XMID\n");
				return 0;
			}
		
			/* Ok now to start part 2 */

			i = len + 8;
		
			/* Not an XMID */
			if (memcmp (stream+i, "CAT ", 4))
			{
				fprintf (stderr, "Not a recognised XMID (%c%c%c%c)\n", stream[i], stream[i+1], stream[i+2], stream[i+3]);
				return 0;	
			}
			i += 4;
		}
		
		len = stream[i+3];
		len |= stream[i+2] << 8;
		len |= stream[i+1] << 16;
		len |= stream[i] << 24;
	
		i+=4;

		/* Not an XMID */
		if (memcmp (stream+i, "XMID", 4))
		{
			fprintf (stderr, "Not a recognised XMID (%c%c%c%c)\n", stream[i], stream[i+1], stream[i+2], stream[i+3]);
			return 0;	
		}
		
		i+=4;

		/* Ok it's an XMID, so pass it to the ExtractCode */

		xmidi->events = (midi_event**)malloc (sizeof(midi_event*) * xmidi->tracks);
		xmidi->timing = (short int*)malloc (sizeof(short) * xmidi->tracks);
		
		for (j = 0; j < xmidi->tracks; j++)
		{
			xmidi->timing[j] = 0;
			xmidi->events[j] = NULL;
		}

		count = extractTracks (xmidi, stream+i, len);

		if (count != xmidi->tracks)
		{
			fprintf (stderr, "Error: unable to extract all (%i) tracks specified from XMidiFile. Only %i read\n", xmidi->tracks, count);
			return 0;
		}

		return 1;
		
	}
	
	return 0;	
}

/*
// Converts and event list to a MTrk
// Returns bytes of the array
// buf can be NULL
*/
uint32 XMidi::convertListToMTrk (unsigned char *buf, midi_event *mlist)
{
	int time = 0;
	midi_event	*event;
	uint32	delta = 0;
	unsigned char	last_status = 0;
	uint32 	i = 8;
	uint32 	j;
	int i_start;

	/* This is set true to make the song end when an XMidiFile break is hit. */
	int	sshock_break = 0;

	if (buf)
	{
		buf[0] = 'M';
		buf[1] = 'T';
		buf[2] = 'r';
		buf[3] = 'k';
	}

	for (event = mlist; event; event=event->next)
	{
		i_start = i;

		/* If sshock_break is set, the delta is only 0 */
		delta = sshock_break?0:event->time - time;
		time = event->time;

		if (buf) i += putVLQ (buf+i, delta);
		else i += putVLQ (NULL, delta);

		if ((event->status != last_status) || (event->status >= 0xF0))
		{
			if (buf) buf[i] = event->status;
			i++;
		}
		
		last_status = event->status;
		
		switch (event->status >> 4)
		{
			/* 2 bytes data */
			/* Note off, Note on, Aftertouch and Pitch Wheel */
			case 0x8: case 0x9: case 0xA: case 0xE:
			if (buf)
			{
				buf[i] = event->data[0];
				buf[i+1] = event->data[1];
			}
			i += 2;
			break;
			
			/* Controller, we need to catch XMIXI Breaks */
			case 0xB:
			if (buf)
			{
				buf[i] = event->data[0];
				buf[i+1] = event->data[1];
			}
			/* XMidiFile Break */
			if (event->data[0] == XMIDI_CONTROLLER_NEXT_BREAK) sshock_break = 1;
			i += 2;
			break;

			/* 1 bytes data */
			/* Program Change and Channel Pressure */
			case 0xC: case 0xD:
			if (buf) buf[i] = event->data[0];
			i++;
			break;
			

			/* Variable length */
			/* SysEx */
			case 0xF:
			if (event->status == 0xFF)
			{
				if (buf) buf[i] = event->data[0];
				i++;
			}
	
			if (buf) i += putVLQ (buf+i, event->len);
			else i += putVLQ (NULL, event->len);
			
			if (event->len)
			{
				for (j = 0; j < event->len; j++)
				{
					if (buf) buf[i] = event->stream[j]; 
					i++;
				}
			}

			break;
			

			/* Never occur */
			default:
			fprintf (stdout, "Not supposed to see this\n");
			break;
		}
	}
	

	if (buf)
	{
		uint32	size = i -8;
		buf[4] = (unsigned char) ((size >> 24) & 0xFF);
		buf[5] = (unsigned char) ((size >> 16) & 0xFF);
		buf[6] = (unsigned char) ((size >> 8) & 0xFF);
		buf[7] = (unsigned char) (size & 0xFF);
	}
	
	return i;
}

void XMidi::deleteEventList (midi_event *mlist)
{
	midi_event *event;
	midi_event *next;
	
	next = mlist;
	event = mlist;

	while ((event = next))
	{
		next = event->next;
		if (event->stream) free(event->stream);
		free(event);
	}
}

